// DO NOT EDIT: Autogenerated by version_generator.py.

#include <string>

namespace mlperf {

const std::string& LoadgenVersion() {
  static const std::string str = "6.0.10";
  return str;
}

const std::string& LoadgenBuildDateLocal() {
  static const std::string str = "2026-02-06T17:40:02.299619";
  return str;
}

const std::string& LoadgenBuildDateUtc() {
  static const std::string str = "2026-02-06T17:40:02.299635";
  return str;
}

const std::string& LoadgenGitRevision() {
  static const std::string str = "0b4aa6f62e";
  return str;
}

const std::string& LoadgenGitCommitDate() {
  static const std::string str = "2026-02-06T17:39:21Z";
  return str;
}

const std::string& LoadgenGitStatus() {
  static const std::string str = R"LGVG_RSLD()LGVG_RSLD";
  return str;
}

const std::string& LoadgenGitLog() {
  static const std::string str = R"LGVG_RSLD(0b4aa6f62e9fc33fd816dd9725b29b6ce65831e2 Increment version to 6.0.10)LGVG_RSLD";
  return str;
}

const std::string& LoadgenSha1OfFiles() {
  static const std::string str = R"LGVG_RSLD(012aad77e5206c89d50718c46c119d1f3cb056b2 /.clang-format
a8ebd64f62d0349aeedbe3295d833ebdce625c2e /CMakeLists.txt
ddeb472d62edf2920db1f8fa3beebe3e831557f1 /MANIFEST.in
e850133bdbbfa62c84bc05a7358114d8996e0530 /README.md
a7e4fcdb5572fa084dcb084f1b4cb658b4b7f38f /README_BUILD.md
01f9ae9887f50bc030dc6107e740f40c43ca388f /README_FAQ.md
4badd8116e0802b47b58337d6c56a00d5a8ad2e8 /VERSION.txt
d013101621ef06a0ddc5e7d9ce511918a8b2ebe6 /__init__.py
14d178b64c7fc45d090e038c08d9b78ca943c383 /bindings/c_api.cc
23d9f99e00b2d196e095fae0bb453a391c18d601 /bindings/c_api.h
90e9c362d9c0e3fdb7d9fad70b34227d7b70f205 /bindings/python_api.cc
53dba8ad4272190ceb6335c12fd25e53dc02a8cb /diagram_network_submission.png
84c2f79309b237cef652aef6a187ba8e875a3952 /diagram_submission.png
0cd7b546a389deac73f7955cd39255ed76557d62 /early_stopping.cc
158fcae6a5f47e82150d6416fa1f7bcef37e77fe /early_stopping.h
02fcfe6d9cf958eeb4b6f1f4dbe87ba7eb4d7dec /issue_query_controller.cc
ed20934fd3507a15949d501ac154be38e766f6ab /issue_query_controller.h
86bc3765857092b9d1f380604b101c771679842a /loadgen.cc
ce9fcb5d44951e7e9048a83b7c1a41c8b8e0f7d8 /loadgen.h
47f748307536f80cfc606947b440dd732afc2637 /loadgen_integration_diagram.svg
49e63158ebca654fa4b7c5f3321054cf4d6c3a30 /logging.cc
2102c91dedbaa156beadf0cecc63d2f43a2bd7dd /logging.h
661c4040c38284469c08cd866a4687107b3d1873 /mlperf.conf
0bdb83f7ad95d931c7f93bf074d042d3b790705a /mlperf_conf.h
712fab87b72ba67ef2a068d0f9f47da65130342f /pyproject.toml
1f18e9cd3ee4dc89a387cf462de1d0ceb1ece975 /query_dispatch_library.h
c4f399103bc3d172079bbd4cd2b0ca0f22eebc4f /query_sample.h
8323a2225be1dff31f08ecc86b76eb3de06568bc /query_sample_library.h
a5ff7e77caa6e9e22ada90f0de0c865c987bf167 /requirements.txt
fde4e173d3ec8461dd439ff0d8e605b52c4becf5 /results.cc
fce22d5a588d91fd968a6b25c27896dba87bc276 /results.h
8dd48aeee1f0cdb30e27d53033bad4989ea11489 /setup.py
18d4809589dae33317d88d9beeb5491a6e1ccdec /system_under_test.h
088be02e337dd5610a991e8beac38cfb0fe49d3a /test_settings.h
1e6b962f1e22450b3e0ced24292230c64d23e20a /test_settings_internal.cc
711a06cde66a3b817556ce3ca05b62ca163220c6 /test_settings_internal.h
3df8fdabf6eaea4697cf25d1dcb89cae88e36efd /utils.cc
40775e32d619ea6356826ae5ea4174c7911f6894 /utils.h
cbec2a5f98f9786c8c3d8b06b3d12df0b6550fa0 /version.cc
9d574baa64424e9c708fcfedd3dbb0b518a65fcc /version.h
9f23d13276194588473120a8a6ecf5a6ed034a23 /version_generator.py)LGVG_RSLD";
  return str;
}

}  // namespace mlperf
